MoveMesh = Object:extend()


local MoveMeshAttachment = Object:extend()


local function sign(p1, p2, p3)
  return (p1[1] - p3[1]) * (p2[2] - p3[2]) - (p2[1] - p3[1]) * (p1[2] - p3[2])
end


local function pointInTriangle(p, verts)
  local b1 = sign(p, verts[1], verts[2]) < 0
  local b2 = sign(p, verts[2], verts[3]) < 0
  local b3 = sign(p, verts[3], verts[1]) < 0

  return (b1 == b2) and (b2 == b3)
end


function MoveMeshAttachment:new(mesh, curidx, x, y)
  self.mesh = mesh
  self.curidx = curidx
  self.pos = {x, y}
end

function projectVerts(verts)
  local t = {}
  for i, v in ipairs(verts) do
    t[#t+1], t[#t+2] = projectPoint(unpack(v))
  end

  return t
end


function circleLineIntersect(v1, v2, c, r)
  local px, py = lust.tools.point_line_pivot(c[1], c[2], v1[1], v1[2], v2[1], v2[2])
  local d = lust.tools.length(px - c[1], py - c[2])
  return d <= r, px, py
end


function MoveMeshAttachment:cast_ray(dx, dy)
  local curidx = self.curidx
  local cur = self.mesh.mesh[curidx]
  local last = {}
  while true do
    if last[curidx] then
      return
    end
    last[curidx] = curidx
    for i = 1, 3 do
      local v1 = cur.verts[i]
      local v2 = cur.verts[i % 3 + 1]


      local px, py = lust.tools.ray_line_intersection(self.pos[1], self.pos[2], dx, dy, v1[1], v1[2], v2[1], v2[2])
      if px then
        if cur.neighbors[i] < 0 then
          return px, py
        elseif not last[cur.neighbors[i]] then
         -- last = curidx
          curidx = cur.neighbors[i]
          cur = self.mesh.mesh[curidx]
          break
        else
        end
      end
    end

  end
end


function MoveMeshAttachment:findCollEdge(visited, cur, pos, radius)
  local m = self.mesh

  if visited[cur] then
    return
  end

  visited[cur] = cur

  for i, n in ipairs(m.mesh[cur].neighbors) do
    local v1 = m.mesh[cur].verts[i]
    local v2 = m.mesh[cur].verts[i % 3 + 1]

    local intersect, ipx, ipy = circleLineIntersect(
            v1,
            v2,
            pos, radius)


    if intersect then
      if n > 0 then
        local c, jpx, jpy = self:findCollEdge(visited, n, pos, radius)
        if c then
          return c, jpx, jpy
        end
      else
        local v = {v1, v2}
        return v, ipx, ipy
      end
    end

  end

end


function MoveMeshAttachment:collision(radius)
  local visitedTris = {}
  local edges = {}

  local p = {self.pos[1], self.pos[2]}

  local found = false
  local colEdge, ipx, ipy = true, 0, 0
  while colEdge do
    colEdge, ipx, ipy = self:findCollEdge(visitedTris, self.curidx, p, radius)
    if colEdge then
      found = true
      local dx, dy = lust.tools.scale(p[1] - ipx, p[2] - ipy, radius)
      p[1] = ipx + dx
      p[2] = ipy + dy
    end
  end

  return found, p[1] - self.pos[1], p[2] - self.pos[2]

end

function MoveMesh:draw()
  love.graphics.setColor(255, 0, 0)
  for i, t in ipairs(self.mesh) do
    local tt = projectVerts(t.verts)
--    love.graphics.polygon('line', tt)
    local cx = 0
    local cy = 0
    for j = 1, 3 do
      local k = j % 3 + 1
      
      cx = cx + tt[2*(j-1)+1]
      cy = cy + tt[2*(j-1)+2]

      if self.visited[i] then
        love.graphics.setLineWidth(5)
      else
        love.graphics.setLineWidth(1)
      end

      love.graphics.line(tt[2*(j-1)+1], tt[2*(j-1)+2], tt[2*(k-1)+1], tt[2*(k-1)+2])
    end
    cx = cx / 3
    cy = cy / 3

    love.graphics.print(i, cx, cy)
  end
  love.graphics.setColor(255, 255, 255)
  self.visited = {}
end


function MoveMeshAttachment:moveto(x, y)
  local m = self.mesh.mesh
  local intri = m[self.curidx]
  local pn = {x, y}
  local inside = pointInTriangle(pn, intri.verts)

  if not inside then
    for i, t in ipairs(intri.neighbors) do
      if t > 0 then
        
        inside = pointInTriangle(pn, m[t].verts)
        if inside then
          self.curidx = t
          break
        end
      end
    end
  end

  if inside then
    self.pos = pn
  end

  return inside
end


local function len(v1, v2)
  local x = v1[1] - v2[1]
  local y = v1[2] - v2[2]
  return math.sqrt(x*x+y*y)
end


local function triangleArea(v1, v2, v3)
  local a = len(v1, v2)
  local b = len(v2, v3)
  local c = len(v3, v1)
  local s = (a+b+c)/2
  return math.sqrt(s*(s-a)*(s-b)*(s-c))
end


local function heightOnTriangle(p, verts)
  local at = triangleArea(verts[1], verts[2], verts[3])
  local a1 = triangleArea(p, verts[2], verts[3])
  local a2 = triangleArea(p, verts[1], verts[3])
  local a3 = triangleArea(p, verts[1], verts[2])

  return (verts[1][3] * a1 + verts[2][3] * a2 + verts[3][3] * a3) / at
end


function MoveMeshAttachment:height()
  return heightOnTriangle(self.pos, self.mesh.mesh[self.curidx].verts)
end


function MoveMesh:new(filename)
  self.mesh = love.filesystem.load(filename)()
  self.visited = {}
end


function MoveMesh:locate(x, y, z)
  local cm = self.mesh
  local p = {x, y}
  -- TODO find highest triangle below obj
  for i, t in ipairs(cm) do
    local inside = pointInTriangle(p, t.verts)
    if inside then
      return MoveMeshAttachment(self, i, x, y)
    end
  end

  --error("Not on the mesh :-(")
end


